from pathlib import Path
import polars as pl
import duckdb
import orjson
from tqdm import tqdm
import subprocess
import os
import sys
import requests
from loguru import logger
from concurrent.futures import ThreadPoolExecutor
from PIL import Image, ImageFile, UnidentifiedImageError
from .function_text import PipelineText
from item_matching.func.utilities import make_dir

logger.remove()
logger.add(sys.stdout, colorize=True, format='<level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')
ImageFile.LOAD_TRUNCATED_IMAGES = True


class PipelineImage:
    def __init__(
            self,
            path: Path,
            mode: str = ''
    ):
        self.path_image = path / 'download_img'
        self.col_img_download = 'image_url'
        self.mode = mode

        # init path image
        make_dir(self.path_image)

    def download_images_request(self):
        def _download(arr: dict):
            # init
            url = arr[self.col_img_download]
            idx = arr['index']

            # get
            response = requests.get(url, stream=True)

            # path
            path_img = folder / f'{idx}.jpg'
            path_json = folder / f'{idx}.json'

            # resize and log json
            if not path_img.exists():
                try:
                    # img
                    img = Image.open(response.raw).convert('RGB').resize((224, 224))
                    img.save(str(path_img))
                    # json
                    json_object = orjson.dumps(arr, option=orjson.OPT_INDENT_2).decode("utf-8")
                    with open(str(path_json), 'w') as outfile:
                        outfile.write(json_object)
                except UnidentifiedImageError as e:
                    pass

        # path
        folder = self.path_image / f'img_{self.mode}/00000'
        if not folder.exists():
            make_dir(folder)
            # read data
            query = f"""select * from read_parquet('{self.path_image}/{self.mode}_0.parquet')"""
            df = (
                duckdb.sql(query).pl()
                .with_row_index()
            )
            # run
            run = df.to_dicts()
            with ThreadPoolExecutor() as executor:
                list(tqdm(executor.map(_download, run), total=len(run)))

    def download_images_img2dataset(self):
        folder = self.path_image / f'img_{self.mode}'
        if not folder.exists():
            os.chdir(str(self.path_image))
            command = (
                f"img2dataset --url_list={self.mode}_0.parquet "
                f"--output_folder=img_{self.mode}/ "
                f"--processes_count=16 "
                f"--thread_count=32 "
                f"--image_size=224 "
                f"--output_format=files "
                f"--input_format=parquet "
                f"--url_col={self.col_img_download} "
                f"--number_sample_per_shard=50000 "
            )
            subprocess.run(command, shell=True)

    def load_images(self) -> pl.DataFrame:
        # listing
        path = self.path_image / f'img_{self.mode}'
        lst_json = sorted(path.glob('*/*.json'))
        lst_file = [orjson.loads(open(str(i), "r").read())['url'] for i in tqdm(lst_json, desc='Loading json in folder')]
        lst_img = [str(i) for i in tqdm(sorted(path.glob('*/*.jpg')), desc='Loading jpg in folder')]
        df = pl.DataFrame({
            f'{self.mode}_{self.col_img_download}': lst_file,
            f'{self.mode}_file_path': lst_img,
            f'{self.mode}_exists': [True] * len(lst_file),
        })

        logger.info(f'[Data] Load Images: {df.shape}')
        return df

    def run(
            self,
            data,
            download: bool = False,
            download_mode: str = 'img2dataset',
            col_img_url: str = 'images',
    ):
        # edit url
        # col_query = f",concat('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) {self.col_image}"
        col_query = f"{col_img_url} {self.col_img_download}"

        # load data
        query = f"""select *, {col_query} from data"""
        df = duckdb.sql(query).pl()
        df.write_parquet(self.path_image / f'{self.mode}_0.parquet')
        logger.info(f'[Data] Base Data {self.mode}: {df.shape}')

        # download
        if download:
            self.download_images_img2dataset() if download_mode == 'img2dataset' else self.download_images_request()

        # load data image
        data_img = self.load_images()

        # errors image
        if data_img.shape[0] == 0:
            logger.info(f'[Data] Images Errors {self.mode}: {data.shape}')
            return data, data_img

        # join
        data = (
            df
            .pipe(PipelineText.clean_text)
            .select(pl.all().name.prefix(f'{self.mode}_'))
            .join(data_img, on=f'{self.mode}_{self.col_img_download}', how='left')
            .filter(pl.col(f'{self.mode}_exists'))
        )
        logger.info(f'[Data] Join Images {self.mode}: {data.shape}')
        return data, data_img



from datasets import Dataset
from time import perf_counter
from autofaiss import build_index
from tqdm import tqdm
import polars as pl
from pymilvus import MilvusClient
import numpy as np
from pathlib import Path
from collections import defaultdict
from data_load import load


# init data
df, col, _ = load()
path = Path('/media/kevin/data_4t/Test/benchmark')
embeddings = np.load(path / 'array/bge.npy')
df = df.with_columns(pl.Series(values=embeddings, name='vector'))


def faiss():
    # index
    start = perf_counter()
    file_index = str(path / 'ip.index')

    build_index(
        embeddings=embeddings,
        index_path=file_index,
        index_infos_path=str(path / f'index.json'),
        save_on_disk=True,
        metric_type='ip',
        verbose=30,
    )
    end_index = perf_counter() - start

    # query
    dataset = Dataset.from_polars(df)
    dataset.load_faiss_index('vector', file_index)

    start = perf_counter()
    total_sample = len(dataset)
    batch_size = 20_000
    num_batches = (total_sample + batch_size) // batch_size
    print(f'Total batches: {num_batches}, Batch size: {batch_size:,.0f}')
    for i, idx in tqdm(enumerate(range(num_batches), start=1), total=num_batches):
        start_idx = idx * batch_size
        end_idx = min(start_idx + batch_size, total_sample)
        batch = embeddings[start_idx:end_idx]

        score, result = dataset.get_nearest_examples_batch(
            'vector',
            batch,
            k=10,
        )
    end_query = perf_counter() - start
    return end_index, end_query


def milvus():
    data = df.to_dicts()

    # index
    client = MilvusClient("milvus_demo.db")
    collection = 'demo_collection'
    if client.has_collection(collection_name=collection):
        client.drop_collection(collection_name=collection)

    client.create_collection(
        collection_name=collection,
        dimension=embeddings.shape[1],
        metric_type="IP"
    )

    start = perf_counter()
    total_sample = len(data)
    batch_size = 500
    num_batches = (total_sample + batch_size) // batch_size
    print(f'Total batches: {num_batches}, Batch size: {batch_size:,.0f}')
    for i, idx in tqdm(enumerate(range(num_batches), start=1), total=num_batches):
        start_idx = idx * batch_size
        end_idx = min(start_idx + batch_size, total_sample)
        batch = data[start_idx:end_idx]
        res = client.insert(collection_name="demo_collection", data=batch)
    end_index = perf_counter() - start

    # query
    start = perf_counter()
    batch_size = 100
    num_batches = (total_sample + batch_size) // batch_size
    print(f'Total batches: {num_batches}, Batch size: {batch_size:,.0f}')
    for i, idx in tqdm(enumerate(range(num_batches), start=1), total=num_batches):
        start_idx = idx * batch_size
        end_idx = min(start_idx + batch_size, total_sample)
        batch = embeddings[start_idx:end_idx].tolist()
        res = client.search(
            collection_name=collection,
            data=batch,
            limit=10,
            search_params={"metric_type": "IP", "params": {}},
            output_fields=col
        )
    end_query = perf_counter() - start
    return end_index, end_query


dict_ = defaultdict(list)
bench = {
    'faiss': faiss,
    'milvus': milvus,
}
for i, f in bench.items():
    end_index, end_query = f()
    dict_['name'].append(i)
    dict_['time_index'].append(end_index)
    dict_['time_query'].append(end_query)

report = (
    pl.DataFrame(dict_)
    .with_columns(pl.lit(f'{embeddings.shape}').alias('data_size'))
)
report.write_csv(path / 'database_benchmark.csv')


import torch
from loguru import logger
from PIL import Image
from torch.nn import functional as F
from transformers import AutoProcessor, SiglipVisionModel
from FlagEmbedding import BGEM3FlagModel

device = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'


class Model:
    def __init__(self):
        self.device = device
        self.text_model = None
        self.img_processor = None
        self.img_model = None
        logger.info(f'[Model] Run on: {self.device}')

    def get_img_model(self):
        """Get Image Model"""
        model_id = 'google/siglip-base-patch16-224'
        self.img_processor = AutoProcessor.from_pretrained(model_id)
        self.img_model = SiglipVisionModel.from_pretrained(model_id).to(self.device)

    def get_text_model(self):
        model_id = 'BAAI/bge-m3'
        self.text_model = BGEM3FlagModel(model_id, use_fp16=True)

    def process_text(self, list_text: list[str]):
        embeddings = self.text_model.encode(
            list_text,
            batch_size=512,
            max_length=80,
            return_dense=True,
            return_sparse=False,
            return_colbert_vecs=False
        )['dense_vecs']
        return embeddings

    def process_image(self, batch, col: str):
        images = [Image.open(i).convert('RGB') for i in batch[col]]
        inputs = self.img_processor(images=images, return_tensors='pt').to(device)
        with torch.inference_mode():
            outputs = self.img_model(**inputs)
        pooled_output = outputs.pooler_output
        return {'image_embed': pooled_output}

    @staticmethod
    def pp_normalize(batch, col: str = 'image_embed') -> dict:
        norm_embed = F.normalize(batch[col], p=2, dim=1).cpu().numpy()
        return {col: norm_embed}



from pydantic import BaseModel, Field, computed_field
import polars as pl
from datasets import Dataset, concatenate_datasets
from pathlib import Path
import numpy as np
from item_matching.model.model import Model
from item_matching.func.utilities import make_dir
from loguru import logger


class ConfigEmbedding(BaseModel):
    ROOT_PATH: Path = Field(default=None)
    SHARD_SIZE: int = Field(default=1_500_000)
    MODE: str = Field(default='')
    MATCH_BY: str = Field(default='text')

    @computed_field
    @property
    def col_embedding(self) -> str:
        dict_ = {'image': 'image_embed'}
        return dict_.get(self.MATCH_BY, 'dense_embed')

    @computed_field
    @property
    def col_input(self) -> str:
        dict_ = {'image': f'{self.MODE}_file_path'}
        return dict_.get(self.MATCH_BY, f'{self.MODE}_item_name_clean')

    @computed_field
    @property
    def path_array(self) -> Path:
        return self.ROOT_PATH / f'{self.MODE}_array'

    @computed_field
    @property
    def path_ds(self) -> Path:
        return self.ROOT_PATH / f'{self.MODE}_ds'


class DataEmbedding:
    def __init__(self, config_input: ConfigEmbedding):
        # Path
        self.config_input = config_input

        # Init path
        make_dir(self.config_input.path_array)
        make_dir(self.config_input.path_ds)

    def load(self, data: pl.DataFrame):
        # Log total chunks
        total_sample = len(data)
        num_chunks = (total_sample + self.config_input.SHARD_SIZE) // self.config_input.SHARD_SIZE

        # Model
        model = Model()
        if self.config_input.MATCH_BY == 'text':
            model.get_text_model()
        elif self.config_input.MATCH_BY == 'image':
            model.get_img_model()
        col_embed = self.config_input.col_embedding

        # Process and save each chunk
        logger.info(
            f'[DataEmbedding] Total chunks {self.config_input.MODE}: {num_chunks} '
            f'- Shard size: {self.config_input.SHARD_SIZE:,.0f}'
        )
        for i, idx in enumerate(range(num_chunks), start=1):
            # Check if exists:
            dataset_name = self.config_input.path_ds / f'{idx}'
            array_name = self.config_input.path_array / f'{idx}.npy'
            if dataset_name.exists():
                continue

            # Load Chunk
            start_idx = idx * self.config_input.SHARD_SIZE
            end_idx = min(start_idx + self.config_input.SHARD_SIZE, total_sample)
            dataset_chunk = Dataset.from_polars(data[start_idx:end_idx])
            logger.info(f'Shard [{i}/{num_chunks}]: start {start_idx:,.0f} end {end_idx:,.0f}')

            # Process dataset
            if self.config_input.MATCH_BY == 'text':
                embeddings = model.process_text(dataset_chunk[self.config_input.col_input])
                dset_embed = Dataset.from_dict({col_embed: embeddings})
                dataset_chunk = concatenate_datasets([dataset_chunk, dset_embed], axis=1)
            else:
                dataset_chunk = dataset_chunk.map(
                    model.process_image,
                    batch_size=512,
                    batched=True,
                    fn_kwargs={'col': self.config_input.col_input}
                )

            # Normalize
            dataset_chunk.set_format(type='torch', columns=[col_embed], output_all_columns=True)
            if self.config_input.MATCH_BY == 'image':
                dataset_chunk.map(Model.pp_normalize, batched=True, fn_kwargs={'col': col_embed})
            dataset_chunk.set_format(type='numpy', columns=[col_embed], output_all_columns=True)

            # Save chunk
            np.save(array_name, dataset_chunk[col_embed])
            dataset_chunk.save_to_disk(str(dataset_name))



from core_pro import Drive
from core_pro.ultilities import make_dir
from pathlib import Path


path = Path.home() / 'Downloads/Data/Item_Matching_Test'
make_dir(path)
drive_id = '1TNgm50QJ3w2-h7cFw5b93cgKAK-zHaY1'
Drive().drive_download(drive_id, str(path))
